"""Core workflow models used across the system."""
import uuid
from datetime import datetime
from enum import Enum
from typing import Literal

from pydantic import BaseModel, Field

from olmoearth_shared.api.common.search_filters import DatetimeFilter, KeywordFilter, SortDirection
from olmoearth_shared.api.run.evaluation_results import EvaluationMetricType
from olmoearth_shared.api.run.status import Status
from olmoearth_shared.api.run.workflow_args import WorkflowArgs, WorkflowCreateArgs
from olmoearth_shared.api.run.workflow_notification_config import WorkflowNotificationConfig
from olmoearth_shared.api.run.workflow_type import WorkflowType


class WandbRunInfo(BaseModel):
    """Information about a Weights & Biases run."""
    run_id: str = Field(description="Weights & Biases run ID")
    url: str = Field(description="URL to the Weights & Biases run")


class FineTuningWorkflowResponse(BaseModel):
    """Response model for completed fine tuning workflows."""
    workflow_type: Literal[WorkflowType.FINE_TUNING] = Field(default=WorkflowType.FINE_TUNING)
    windows_count: int = Field(description="Number of labeled windows created during the fine-tuning workflow")
    model_pipeline_id: uuid.UUID = Field(description="The ID of the pipeline housing the newly fine-tuned model stage")
    wandb_run_info: WandbRunInfo = Field(description="Weights & Biases run information")
    metric_types: list[EvaluationMetricType] = Field(default_factory=list, description="Evaluation metric types computed (e.g., ['confusion_matrix'])")


class DatasetBuildFromWindowsWorkflowResult(BaseModel):
    """Result model for completed dataset build from windows workflows."""
    workflow_type: Literal[WorkflowType.DATASET_BUILD_FROM_WINDOWS] = Field(default=WorkflowType.DATASET_BUILD_FROM_WINDOWS)


class WorkflowMetrics(BaseModel):
    """
    This tracks some interesting metrics at the Workflow level.
    """

    request_area_sq_km: float | None = Field(default=None, description="The area of the request in square kilometers.")
    dataset_size_mb: float | None = Field(default=None, description="The size of the dataset in megabytes")
    total_compute_time_seconds: float | None = Field(default=None, description="The total compute time in seconds for this workflow")


class WorkflowResponse(BaseModel):
    """
    The model we provide when asked for a Workflow over the API
    """

    id: uuid.UUID
    args: WorkflowArgs
    external_id: str | None = Field(
        default=None,
        description="The provided external ID of the Workflow, if applicable.",
    )
    progress: int = Field(ge=0, le=100, description="The progress of the Workflow, from 0 to 100")
    status: Status = Field(description="The status of the Workflow")
    metrics: WorkflowMetrics = Field(description="Metrics related to the Workflow's execution")
    created_at: datetime
    updated_at: datetime
    completed_at: datetime | None = Field(
        default=None,
        description="The date and time the Workflow was completed. Could be successful or failure",
    )
    error_message: str | None = Field(
        default=None,
        description="Error details if the Workflow failed",
    )
    notification_config: WorkflowNotificationConfig | None = Field(
        default=None,
        description="Notification configuration for this workflow, if any",
    )


class WorkflowCreateRequest(BaseModel):
    """
    The model we expect when creating a new Workflow.
    """

    args: WorkflowCreateArgs = Field(description="The arguments for the Workflow")
    external_id: str | None = Field(
        default=None,
        description="An optional external ID for the Workflow, useful for tracking in external systems.",
    )
    notification_config: WorkflowNotificationConfig | None = Field(
        default=None,
        description="Optional configuration for notifications related to this Workflow.",
    )


class WorkflowSortField(str, Enum):
    """Valid fields for sorting workflows in search results."""

    STATUS = "status"
    CREATED_AT = "created_at"
    UPDATED_AT = "updated_at"
    COMPLETED_AT = "completed_at"
    PROGRESS = "progress"


class SearchWorkflowsRequest(BaseModel):
    """Request model for searching workflows"""

    id: KeywordFilter[uuid.UUID] | None = None
    external_id: KeywordFilter[str] | None = None
    status: KeywordFilter[Status] | None = None
    created_at: DatetimeFilter | None = None
    updated_at: DatetimeFilter | None = None
    completed_at: DatetimeFilter | None = None

    # Pagination and sorting
    limit: int = Field(default=50, ge=1, le=1000)
    offset: int = Field(default=0, ge=0)
    sort_by: WorkflowSortField = Field(default=WorkflowSortField.CREATED_AT)
    sort_direction: SortDirection = Field(default=SortDirection.DESC)



"""
Point-to-pixel window preparer implementation.

This module provides a concrete implementation of VectorFeaturesWindowPreparer
that creates single-pixel windows from point annotations. Each point annotation
becomes a 1x1 pixel window centered on the point location fit for training a classifier.
"""

from rslearn.utils import STGeometry, get_utm_ups_crs
from shapely.geometry import Point, box

from olmoearth_run.runner.models.training.labeled_data import (
    AnnotationTask,
    LabeledSTGeometry,
    LabeledWindow,
)
from olmoearth_run.runner.tools.labeled_window_preparers.labeled_window_preparer import (
    VectorLabelsWindowPreparer,
)
from olmoearth_run.runner.tools.labeled_window_preparers.geometry_utils import (
    project_geometry_to_crs,
)


class PointToPixelWindowPreparer(VectorLabelsWindowPreparer):
    """
    Window preparer that creates single-pixel windows from point annotations.

    Each point annotation becomes a 1x1 pixel window centered on the point location.
    This is useful for point-based classification tasks where each annotation
    represents a single pixel to be classified.
    """

    def __init__(self, window_resolution: float = 10.0):
        """
        Initialize the point-to-pixel window preparer.

        Args:
            window_resolution: Resolution in meters per pixel (default: 10.0 for Helios)
        """
        self.window_resolution = window_resolution

    def prepare_labeled_windows(
        self,
        annotation_task: AnnotationTask,
    ) -> list[LabeledWindow[list[LabeledSTGeometry]]]:
        """
        Prepare labeled windows with vector labels from point annotations.

        For each point annotation, creates a single-pixel window centered on the point.
        The window is sized to be exactly 1x1 pixels at the specified resolution.

        Args:
            annotation_task: Single AnnotationTask object containing task context and annotations

        Returns:
            List of LabeledWindow objects with single-pixel windows and point labels
        """
        # Skip tasks with no annotations - can't create meaningful training windows
        if not annotation_task.annotations:
            return []

        labeled_windows = []

        # Process each annotation in the task
        for annotation_idx, annotation in enumerate(annotation_task.annotations):
            # Create a single-pixel window for this point annotation
            labeled_window = self._create_single_pixel_window(
                annotation, annotation_task, annotation_idx
            )
            labeled_windows.append(labeled_window)

        return labeled_windows

    def _create_single_pixel_window(
        self, annotation: LabeledSTGeometry, task: AnnotationTask, annotation_idx: int
    ) -> LabeledWindow[list[LabeledSTGeometry]]:
        """
        Create a single-pixel window from a point annotation.

        Args:
            annotation: The point annotation to create a window for
            task: The parent annotation task
            annotation_idx: The index of this annotation within the task

        Returns:
            LabeledWindow with a single-pixel window and the point label
        """
        # Extract the point geometry
        point_geom = annotation.st_geometry.shp
        if not isinstance(point_geom, Point):
            raise ValueError(f"Expected Point geometry, got {type(point_geom)}")

        # Calculate CRS based on point location (not task centroid)
        # Since each point creates an independent window, use the point's coordinates
        # for proper UTM/UPS zone selection - this handles large tasks spanning multiple zones
        utm_crs = get_utm_ups_crs(point_geom.x, point_geom.y)

        # Convert to appropriate projection if needed
        projected_geometry = project_geometry_to_crs(
            point_geom, self.window_resolution, utm_crs
        )

        # Create a 1x1 pixel window centered on the point
        window_bounds = self._compute_single_pixel_bounds(projected_geometry)

        # Create a window geometry that matches the computed bounds
        window_geometry = box(*window_bounds)

        # Create the labeled window with the projected point annotation as the label
        # We need to project the annotation to the same coordinate system as the window
        # and use integer pixel coordinates to match the window bounds
        projected_point = projected_geometry.shp
        if not isinstance(projected_point, Point):
            raise ValueError(f"Expected Point geometry, got {type(projected_point)}")
        pixel_point = Point(int(projected_point.x), int(projected_point.y))
        pixel_projected_geometry = STGeometry(
            projected_geometry.projection, pixel_point, projected_geometry.time_range
        )

        projected_annotation = LabeledSTGeometry(
            st_geometry=pixel_projected_geometry, labels=annotation.labels
        )

        return LabeledWindow(
            name=f"task_{task.task_id}_annotation_{annotation_idx}_point_{point_geom.x}_{point_geom.y}",
            st_geometry=STGeometry(
                projected_geometry.projection,
                window_geometry,
                task.task_st_geometry.time_range,
            ),
            labels=[projected_annotation],  # Projected point annotation as the label
        )

    def _compute_single_pixel_bounds(
        self, projected_geometry: STGeometry
    ) -> tuple[int, int, int, int]:
        """
        Compute bounds for a single-pixel window centered on the point.

        This matches the reference implementation's approach:
        - Convert coordinates to integers (pixel coordinates)
        - Create a 1x1 pixel window with no padding
        - Upper bounds are exclusive, so add 1 to maxx and maxy

        Args:
            projected_geometry: The projected point geometry

        Returns:
            Tuple of (minx, miny, maxx, maxy) in pixel coordinates
        """
        point = projected_geometry.shp
        if not isinstance(point, Point):
            raise ValueError(f"Expected Point geometry, got {type(point)}")

        # Convert to integer pixel coordinates
        x, y = int(point.x), int(point.y)

        # Create a 1x1 pixel window with no padding
        minx = x
        miny = y
        maxx = x + 1
        maxy = y + 1

        return (minx, miny, maxx, maxy)



import logging
from abc import ABC, abstractmethod
from typing import Generic

from olmoearth_run.shared.models.api.task_args import (
    TaskArgsType,
)
from olmoearth_run.shared.models.api.task_results import (
    TaskResultsType,
)

logger = logging.getLogger(__name__)

class BaseStepDefinition(ABC, Generic[TaskArgsType, TaskResultsType]):

    @abstractmethod
    def run(self, task_args: TaskArgsType) -> TaskResultsType:
        """A StepDefinition is executed by calling this function"""
        ...

    def on_task_error(self, task_args: TaskArgsType, exc: Exception) -> None:
        """
        This function can be used to do anything on an error after the Task status is updated but before the Task stops.
        """
        pass



from typing import assert_never, cast

from olmoearth_run.runner.models import rslearn_config
from olmoearth_run.runner.models.operational_context import ModelOperationalContext
from olmoearth_shared.models.olmoearth_config.data.output import SegmentationField
from olmoearth_shared.models.olmoearth_config.model.decoders import SegmentationDecoder, UnetSegmentationDecoder, UpsampleConvSegmentationDecoder
from olmoearth_shared.models.olmoearth_config.model.encoders import Encoder, EscapeHatchEncoder
from olmoearth_shared.models.olmoearth_config.model.encoders import OlmoEarthEncoder, OlmoEarthFromHuggingFace, OlmoEarthFromRegistry, OlmoEarthFromModelPath, OlmoEarthFromDistributedCheckpoint
from olmoearth_shared.models.olmoearth_config.model.tasks import ModelTask, SegmentationTask
from olmoearth_shared.models.olmoearth_config.olmoearth_config import OlmoEarthConfig
from olmoearth_shared.models.olmoearth_config.training.optimizers import AdamW
from olmoearth_shared.models.olmoearth_config.training.schedulers import Plateau


class LightningModuleTranspiler:
    """Responsible for transpiling unified OlmoEarthConfig into the top-level `model` member rslearn in model config."""

    @staticmethod
    def generate_lightning_module_config(olmoearth_config: OlmoEarthConfig, ops_context: ModelOperationalContext) -> rslearn_config.LightningModule:
        if not olmoearth_config.model:
            raise ValueError("No model configuration provided, cannot make a model stanza")

        encoder = olmoearth_config.model.encoder
        tasks = olmoearth_config.model.tasks

        if len(tasks) != 1:
            raise NotImplementedError("Only one task is supported for now")

        task_name = list(tasks.keys())[0]
        task = tasks[task_name]

        return rslearn_config.RslearnLightningModule(
            class_path=rslearn_config.ClassPath.RSLEARN_LIGHTNING_MODULE,
            init_args=rslearn_config.RslearnLightningModuleInitArgs(
                model=rslearn_config.Model(
                    class_path=rslearn_config.ClassPath.SINGLE_TASK_MODEL,
                    init_args=rslearn_config.ModelInitArgs(
                        encoder=EncoderTranspilers.generate_encoder_layers(encoder, ops_context),
                        decoder=DecoderTranspilers.generate_decoder_layers(task_name, task, olmoearth_config, ops_context)
                    )
                ),
                scheduler=SchedulerTranspilers.generate_scheduler(olmoearth_config, ops_context),
                optimizer=OptimizerTranspilers.generate_optimizer(olmoearth_config, ops_context)
            )
        )


class EncoderTranspilers:
    """Transpiles model encoder configurations into `encoder` layer specifications for rslearn model config."""

    @staticmethod
    def generate_encoder_layers(encoder: Encoder, ops_context: ModelOperationalContext) -> list[rslearn_config.Encoder]:
        """Delegates to the appropriate encoder generation function based on the encoder type."""
        match encoder:
            case OlmoEarthEncoder():
                return EncoderTranspilers.generate_olmoearth_encoder_layers(encoder, ops_context)
            case EscapeHatchEncoder():
                raise NotImplementedError("Escape hatch encoders are not supported at this time")
            case _ as unreachable:
                assert_never(unreachable)

    @staticmethod
    def generate_olmoearth_encoder_layers(encoder: OlmoEarthEncoder, ops_context: ModelOperationalContext) -> list[rslearn_config.Encoder]:
        """Transpiles OlmoEarth encoder configuration into `encoder` layers for rslearn model config."""
        path_size = encoder.patch_size
        model_id = None
        model_path = None
        checkpoint_path = None
        embedding_size = get_embedding_size(encoder)

        match encoder.source:
            case OlmoEarthFromHuggingFace():
                model_id = encoder.source.model_id
            case OlmoEarthFromRegistry():
                model_path = OpsUtils.get_foundation_model_path(ops_context)
            case OlmoEarthFromModelPath():
                model_path = encoder.source.model_path
            case OlmoEarthFromDistributedCheckpoint():
                checkpoint_path = encoder.source.checkpoint_path
            case _ as unreachable:
                assert_never(unreachable)

        return [rslearn_config.OlmoEarthEncoder(
            class_path=rslearn_config.ClassPath.OLMOEARTH_PRETRAIN_MODEL,
            init_args=rslearn_config.OlmoEarthEncoderInitArgs(
                patch_size=path_size,
                model_id=model_id,
                model_path=model_path,
                checkpoint_path=checkpoint_path,
                embedding_size=embedding_size,
            )
        )]


class DecoderTranspilers:
    """Transpiles model task configuration into the `decoder` layers for the rslearn model architecture spec."""

    @staticmethod
    def generate_decoder_layers(task_name: str, task: ModelTask, olmoearth_config: OlmoEarthConfig, ops_context: ModelOperationalContext) -> list[rslearn_config.Decoder]:
        """Delegates to the appropriate decoder generation function based on the task type."""
        match task:
            case SegmentationTask():
                return DecoderTranspilers.generate_segmentation_decoder_layers(task_name, task.decoder, olmoearth_config, ops_context)
            case _ as unreachable:
                assert_never(unreachable)

    @staticmethod
    def generate_segmentation_decoder_layers(task_name: str, decoder: SegmentationDecoder, olmoearth_config: OlmoEarthConfig, ops_context: ModelOperationalContext) -> list[rslearn_config.Decoder]:
        """Delegates to the appropriate segmentation decoder generation function based on the specific decoder type."""
        match decoder:
            case UpsampleConvSegmentationDecoder():
                return DecoderTranspilers.generate_upsample_conv_segmentation_decoder_layers(olmoearth_config, ops_context)
            case UnetSegmentationDecoder():
                return DecoderTranspilers.generate_unet_segmentation_decoder_layers(task_name, olmoearth_config, ops_context)
            case _ as unreachable:
                assert_never(unreachable)

    @staticmethod
    def generate_upsample_conv_segmentation_decoder_layers(olmoearth_config: OlmoEarthConfig, ops_context: ModelOperationalContext) -> list[rslearn_config.Decoder]:
        # TODO: required for Nov4 release LULC models, which use this decoder architecture. Favyen made some changes to rslearn though, so wait till he updates the olmoearth_projects configs.
        raise NotImplementedError("UpsampleConvSegmentationDecoder is not supported at this time")

    @staticmethod
    def generate_unet_segmentation_decoder_layers(task_name: str, olmoearth_config: OlmoEarthConfig, ops_context: ModelOperationalContext) -> list[rslearn_config.Decoder]:
        """Decoder that processes encoder-produced embeddings through U-Net convolutional layers before passing to the segmentation head."""
        output = olmoearth_config.data.output
        model = olmoearth_config.model

        if not output:
            raise ValueError("No output schema specified, cannot generate U-Net decoder stanza")
        if not model:
            raise ValueError("No model specified, cannot generate U-Net decoder stanza")

        num_classes = len(cast(SegmentationField, output.fields[task_name]).allowed_values)
        encoder = model.encoder

        match encoder:
            case OlmoEarthEncoder():
                patch_size = encoder.patch_size
                encoder_embedding_size = get_embedding_size(encoder)
            case EscapeHatchEncoder():
                raise NotImplementedError("Escape hatch encoders are not supported at this time")
            case _ as unreachable:
                assert_never(unreachable)


        return [
            rslearn_config.UNetDecoder(
                class_path=rslearn_config.ClassPath.UNET_DECODER,
                init_args=rslearn_config.UNetDecoderInitArgs(
                    in_channels=[[patch_size, encoder_embedding_size]],
                    out_channels=num_classes,
                    conv_layers_per_resolution=1,
                )
            ),
            rslearn_config.SegmentationHead(class_path=rslearn_config.ClassPath.SEGMENTATION_HEAD)
        ]


class SchedulerTranspilers:
    @staticmethod
    def generate_scheduler(olmoearth_config: OlmoEarthConfig, ops_context: ModelOperationalContext) -> rslearn_config.Scheduler | None:
        if not ops_context.is_training:
            return None

        training = olmoearth_config.training
        if not training:
            raise ValueError("No training configuration provided, cannot generate scheduler stanza")

        match training.scheduler:
            case Plateau():
                return SchedulerTranspilers.generate_plateau_scheduler(training.scheduler)
            case _ as unreachable:
                assert_never(unreachable)

    @staticmethod
    def generate_plateau_scheduler(scheduler: Plateau) -> rslearn_config.PlateauScheduler:
        return rslearn_config.PlateauScheduler(
            class_path=rslearn_config.ClassPath.PLATEAU_SCHEDULER,
            init_args=rslearn_config.PlateauSchedulerInitArgs(
                factor=scheduler.factor,
                patience=scheduler.patience,
                min_lr=scheduler.min_lr,
                cooldown=scheduler.cooldown,
            )
        )


class OptimizerTranspilers:
    @staticmethod
    def generate_optimizer(olmoearth_config: OlmoEarthConfig, ops_context: ModelOperationalContext) -> rslearn_config.Optimizer | None:
        if not ops_context.is_training:
            return None

        training = olmoearth_config.training
        if not training:
            raise ValueError("No training configuration provided, cannot generate optimizer stanza")

        match training.optimizer:
            case AdamW():
                return OptimizerTranspilers.generate_adamw_optimizer(training.optimizer)
            case _ as unreachable:
                assert_never(unreachable)

    @staticmethod
    def generate_adamw_optimizer(optimizer: AdamW) -> rslearn_config.Optimizer:
        return rslearn_config.AdamWOptimizer(
            class_path=rslearn_config.ClassPath.ADAMW_OPTIMIZER,
            init_args=rslearn_config.AdamWOptimizerInitArgs(
                lr=optimizer.lr,
            )
        )


def get_embedding_size(encoder: OlmoEarthEncoder) -> int:
    match encoder.source:
        case OlmoEarthFromHuggingFace(model_id=model_id):
            # Inline to avoid importing torch unless necessary
            # TODO: maybe move these light declarations into constant files?
            from olmoearth_pretrain.model_loader import ModelID  # type: ignore[import-untyped]
            from rslearn.models.olmoearth_pretrain.model import EMBEDDING_SIZES
            try:
                return EMBEDDING_SIZES[ModelID[model_id]]
            except KeyError as e:
                raise ValueError(f"Model ID {model_id} not found") from e
        case OlmoEarthFromRegistry(embedding_size=embedding_size):
            return embedding_size
        case OlmoEarthFromModelPath(embedding_size=embedding_size):
            return embedding_size
        case OlmoEarthFromDistributedCheckpoint(embedding_size=embedding_size):
            return embedding_size
        case _ as unreachable:
            assert_never(unreachable)


class OpsUtils:
    @staticmethod
    def get_foundation_model_path(ops_context: ModelOperationalContext) -> str:
        return ops_context.model_stage_paths.extra_model_files_path



import logging

import yaml
from rslearn.config import dataset
from upath import UPath

from olmoearth_run.runner.models.operational_context import DataSource, DatasetOperationalContext, ModelOperationalContext
from olmoearth_run.runner.models import rslearn_config
from olmoearth_run.runner.tools.olmoearth_config.loaders.config import ConfigClassLoader
from olmoearth_run.runner.tools.olmoearth_config.transpilers.dataset.label_output_layers import LabelOutputLayers
from olmoearth_run.runner.tools.olmoearth_config.transpilers.dataset.modality_layers import ModalityLayers
from olmoearth_run.runner.tools.olmoearth_config.transpilers.model.data import DataTranspiler
from olmoearth_run.runner.tools.olmoearth_config.transpilers.model.lightning_module import LightningModuleTranspiler
from olmoearth_run.runner.tools.olmoearth_config.transpilers.model.trainer import TrainerTranspiler
from olmoearth_run.shared.models.model_stage_paths import DATASET_CONFIG_FILE_NAME_IN_DATASET_DIR
from olmoearth_shared.models.olmoearth_config.olmoearth_config import OlmoEarthConfig

logger = logging.getLogger(__name__)


def to_dataset_config(olmoearth_config: OlmoEarthConfig, ops_context: DatasetOperationalContext) -> dataset.DatasetConfig:
    """Converts OlmoEarthConfig to rslearn's dataset config object."""
    return dataset.DatasetConfig(
        layers={
            **ModalityLayers.generate_modality_layers(olmoearth_config, ops_context),
            **LabelOutputLayers.generate_label_and_output_layers(olmoearth_config, ops_context)
        },
    )


def to_dataset_config_json(olmoearth_config: OlmoEarthConfig, ops_context: DatasetOperationalContext) -> str:
    """Converts OlmoEarthConfig to rslearn's dataset config as a JSON string."""
    config = to_dataset_config(olmoearth_config, ops_context)
    return config.model_dump_json(exclude_defaults=True, indent=2)


def to_model_config(olmoearth_config: OlmoEarthConfig, ops_context: ModelOperationalContext) -> rslearn_config.ModelYamlConfig:
    """Converts OlmoEarthConfig to rslearn's model config object."""
    return rslearn_config.ModelYamlConfig(
        model=LightningModuleTranspiler.generate_lightning_module_config(olmoearth_config, ops_context),
        data=DataTranspiler.generate_data_module_config(olmoearth_config, ops_context),
        trainer=TrainerTranspiler.generate_trainer_config(olmoearth_config, ops_context),
    )


def to_model_config_yaml(olmoearth_config: OlmoEarthConfig, ops_context: ModelOperationalContext) -> str:
    """Converts OlmoEarthConfig to rslearn's model config as a YAML string."""
    config = to_model_config(olmoearth_config, ops_context)
    return yaml.dump(config.model_dump(mode="json", exclude_defaults=True), default_flow_style=False, sort_keys=False)


def transpile_and_write_dataset_config(olmoearth_config_path: str, dataset_path: UPath) -> None:
    """Transpile OlmoEarthConfig to rslearn DatasetConfig and write to dataset_path/config.json."""
    logger.info(f"Loading OlmoEarthConfig from {olmoearth_config_path}")
    config = ConfigClassLoader.load_olmoearth_config(olmoearth_config_path)

    ops_context = DatasetOperationalContext(
        data_source=DataSource.PLANETARY_COMPUTER,
        dataset_path=str(dataset_path),
    )

    config_path = dataset_path / DATASET_CONFIG_FILE_NAME_IN_DATASET_DIR
    logger.info(f"Writing transpiled dataset config to {config_path}")
    config_path.write_text(to_dataset_config_json(config, ops_context))


def transpile_and_write_model_config(olmoearth_config: OlmoEarthConfig, ops_context: ModelOperationalContext, output_path: UPath) -> None:
    """Transpile OlmoEarthConfig to rslearn model.yaml and write to output_path."""
    logger.info(f"Writing transpiled model config to {output_path}")
    output_path.write_text(to_model_config_yaml(olmoearth_config, ops_context))



import os
from typing import Literal

from pydantic import computed_field
from pydantic_settings import BaseSettings


class Settings(BaseSettings):
    """OlmoEarth Settings, configured via environment variables"""

    # Redis configuration
    REDIS_HOST: str = "localhost"
    REDIS_PORT: int = 6379
    REDIS_USER: str | None = None
    REDIS_PASSWORD: str | None = None
    REDIS_DB: int = 0
    REDIS_MAX_CONNECTIONS: int = 20

    # The number of workers that should be employed to run tasks. Defaults to the number of CPUs available.
    NUM_WORKERS: int = os.cpu_count() or 1

    # Task status database sync configuration
    TASK_STATUS_SYNC_INTERVAL_SECONDS: int = 300  # 5 minutes
    TASK_STATUS_STALENESS_THRESHOLD_MINUTES: int = 15
    SYNC_JOB_REDIS_KEY: str = "olmoearth_run_sync_task_statuses_running"

    # Bulk gcs transfer config
    GCS_LS_PAGE_SIZE: int = 5000

    # The maximum degree of parallelism allowed when processing partitions. If a workflow creates more partitions
    # than this number, multiple partitions will be assigned to a single task and processed serially.
    MAX_PARTITION_PARALLELISM: int = 120

    # The number of dataset build processes that we will launch per CPU core on the machine.
    # Since dataset build is mostly IO bound, this can be set higher than 1.
    DATASET_BUILD_WORKERS_PER_CPU: int = 4

    # The number of pytorch data loaders we will launch per CPU core when running the fit step.
    # Since the data loaders are mostly IO bound, this can be set higher than 1.
    FINE_TUNE_FIT_DATA_LOADERS_PER_CPU: int = 2

    # The URL that executor workers should use to call the API
    # This environment variable is required except for local runner.
    OERUN_API_URL: str = ""

    # Weights and Biases (https://wandb.ai)
    WANDB_API_KEY: str | None = None
    WANDB_PROJECT: str = "olmoearth_run_develop"
    WANDB_ENTITY: str = "eai-ai2"
    WANDB_API_KEY_SECRET_PATH: str = ""

    # PostgreSQL configuration
    POSTGRES_USER: str | None = None
    POSTGRES_PASSWORD: str | None = None
    POSTGRES_HOST: str = "localhost"
    POSTGRES_PORT: int = 5432
    POSTGRES_DB_NAME: str = "esrun"  # legacy name

    GOOGLE_CLOUD_BUCKET_NAME: str | None = None
    GOOGLE_CLOUD_PROJECT_ID: str | None = None
    GOOGLE_CLOUD_REGION: str = "us-west1"
    GOOGLE_CLOUD_SERVICE_ACCOUNT_EMAIL: str = ""

    # Grafana Alloy configuration for runner metrics collection
    OERUNNER_ALLOY_CONFIG_BUCKET_NAME: str = ""
    OERUNNER_ALLOY_CONFIG_OBJECT_KEY: str = ""

    # Grafana Cloud Prometheus remote write configuration
    GRAFANA_CLOUD_PROMETHEUS_URL: str = ""
    GRAFANA_CLOUD_PROMETHEUS_USERNAME: str = ""
    GRAFANA_CLOUD_ACCESS_TOKEN_SECRET_PATH: str = ""

    # Prometheus configuration
    PROMETHEUS_MULTIPROC_DIR: str | None = None  # None if not set (single-process mode)
    PROMETHEUS_METRICS_SERVER_PORT: int = 9090

    # The timeout in hours that we send to google batch.
    MAX_TASK_RUN_DURATION_HOURS: float = 8
    MAX_FINE_TUNING_TASK_DURATION_HOURS: float = 120

    # Google Batch provisioning model: FLEX_START or STANDARD
    GOOGLE_BATCH_PROVISIONING_MODEL: Literal["FLEX_START", "STANDARD"] = "FLEX_START"

    # ElasticSearch configuration
    ELASTIC_HOST: str = "localhost"  # Used in local dev & CI.
    ELASTIC_API_KEY: str | None = None  # Can be empty in local dev & CI
    ELASTIC_FEATURES_INDEX_NAME: str = "features"

    # Admin UI configuration
    ADMIN_UI_DIR: str = "admin_ui/"
    ADMIN_UI_PASSWORD: str | None = None

    # These are GCP secret paths. eg: projects/{project_id}/secrets/{SECRET_NAME}/versions/{version}
    AWS_ACCESS_KEY_ID_SECRET_PATH: str = ""
    AWS_ACCESS_KEY_SECRET_PATH: str = ""

    @computed_field  # type: ignore[prop-decorator]
    @property
    def gcs_storage_manager_thread_count(self) -> int:
        return self.NUM_WORKERS * 2

    @computed_field  # type: ignore[prop-decorator]
    @property
    def google_cloud_temp_prefix(self) -> str:
        return f"gs://{self.GOOGLE_CLOUD_BUCKET_NAME}/temp/"


OlmoEarthSettings = Settings()



import logging
from typing import cast

from rslearn.utils import Projection, STGeometry, get_utm_ups_crs
from rslearn.utils.geometry import WGS84_PROJECTION
from shapely.geometry.base import BaseGeometry

from olmoearth_shared.api.run.prediction_geometry import PredictionRequestCollection, PredictionRequestFeature

logger = logging.getLogger(__name__)


class PartitionInterface:
    """
    A Partitioner is responsible for taking one or more input Geometries and splitting them up into another set of
    (typically smaller) partitions.   This is used in two places during inference:
    1. First to take the user-provided input geometries and split them into partitions that we can hand out to
       different compute nodes for processing.  The number of partitions created determines the maximum concurrency
       of inference.
    2. Second on each compute node; we will take the input partitions and further split them down into
       (typically smaller) windows that we will actually run inference on.  This second step can be optimized
       for whatever the model works best on.
    """

    def __init__(self, output_projection: Projection | None = None, use_utm: bool = False):
        self.output_projection = output_projection if output_projection else WGS84_PROJECTION
        self.use_utm = use_utm

    def partition_request_geometry(self, request_extents: PredictionRequestCollection) -> list[PredictionRequestCollection]:
        """
        Implement this method to describe the specific methodology to partition the input Feature Collection into groups for processing.
        """
        raise NotImplementedError

    def prepare_window_geometries(self, request_extents: PredictionRequestCollection) -> list[tuple[PredictionRequestFeature, STGeometry]]:
        """
        Implement this method to describe the specific methodology to partition input geometries.
        """
        raise NotImplementedError

    def _convert_to_utm(self, geometry: STGeometry) -> STGeometry:
        logger.debug(f"Converting {geometry} to UTM.")
        wgs84_geom = geometry.to_projection(WGS84_PROJECTION)
        wgs84_point = cast(BaseGeometry, wgs84_geom.shp).centroid
        utm_crs = get_utm_ups_crs(wgs84_point.x, wgs84_point.y)
        utm_projection = Projection(utm_crs, self.output_projection.x_resolution, self.output_projection.y_resolution)
        return geometry.to_projection(utm_projection)

    def _update_projection(self, geometry: STGeometry) -> STGeometry:
        if self.use_utm:
            return self._convert_to_utm(geometry)
        else:
            return geometry.to_projection(self.output_projection)



"""Utility functions for geometry operations."""

from collections import defaultdict
from datetime import datetime
from typing import cast

import numpy as np
from rslearn.utils import STGeometry, Projection
from rslearn.utils.geometry import WGS84_PROJECTION
from shapely.geometry import Point, box
from shapely.geometry.base import BaseGeometry

from olmoearth_run.runner.models.training.labeled_data import (
    LabeledSTGeometry,
    RasterLabel,
)
from olmoearth_run.runner.tools.labeled_window_preparers.rasterization_utils import (
    DEFAULT_NODATA_VALUE,
    rasterize_shapes_to_mask,
    transform_geometries_to_pixel_coordinates,
)


def create_window_geometry_from_point(
    projected_point: STGeometry,
    window_buffer: int,
    time_range: tuple[datetime, datetime] | None,
) -> STGeometry:
    """Create a square window geometry centered on the projected point.

    Args:
        projected_point: The projected point geometry
        window_buffer: Buffer around the point in pixels
        time_range: Time range for the window

    Returns:
        STGeometry representing the window bounds
    """
    point = cast(Point, projected_point.shp)

    # Create square polygon centered on the point using box
    minx = point.x - window_buffer
    miny = point.y - window_buffer
    maxx = point.x + window_buffer + 1
    maxy = point.y + window_buffer + 1

    window_polygon = box(minx, miny, maxx, maxy)

    return STGeometry(projected_point.projection, window_polygon, time_range)


def compute_window_bounds(window_geometry: STGeometry) -> tuple[int, int, int, int]:
    """Compute integer bounds for the window from the window geometry.

    Args:
        window_geometry: The window geometry

    Returns:
        Tuple of (minx, miny, maxx, maxy) in world coordinates
    """
    bounds = cast(BaseGeometry, window_geometry.shp).bounds

    minx = int(bounds[0])
    miny = int(bounds[1])
    maxx = int(bounds[2])
    maxy = int(bounds[3])

    return (minx, miny, maxx, maxy)


def create_raster_labels_from_annotations(
    annotations: list[LabeledSTGeometry],
    window_bounds: tuple[int, int, int, int],
    window_resolution: float,
    crs: str,
    dtype: type[np.generic] = np.uint8,
    nodata_value: int = DEFAULT_NODATA_VALUE,
) -> list[RasterLabel]:
    """Create raster labels by rasterizing annotations.

    Args:
        annotations: List of labeled annotations to rasterize
        window_bounds: Window bounds in world coordinates (minx, miny, maxx, maxy)
        window_resolution: Resolution in meters per pixel
        crs: The target coordinate reference system
        nodata_value: Value to use for nodata pixels
        dtype: Data type of the output array

    Returns:
        List of RasterLabel objects with rasterized labels
    """
    if not annotations:
        return []

    # Collect labeled shapes for rasterization
    keys = sorted(annotations[0].labels.keys())
    shape_labels: dict[str, list[tuple[BaseGeometry, int | float]]] = defaultdict(list)

    for annotation in annotations:
        # Project the annotation geometry to the specified CRS
        projected_annotation = project_geometry_to_crs(
            cast(BaseGeometry, annotation.st_geometry.shp), window_resolution, crs
        )

        # Add to shape labels with the label values
        for key in keys:
            label_value = annotation.labels[key]
            if label_value is not None:
                shape_labels[key].append(
                    (cast(BaseGeometry, projected_annotation.shp), label_value)
                )

    # Calculate the expected raster dimensions from the window bounds
    height = window_bounds[3] - window_bounds[1]  # maxy - miny
    width = window_bounds[2] - window_bounds[0]  # maxx - minx

    raster_labels: list[RasterLabel] = []
    for key in keys:
        # Transform geometries to pixel coordinates and rasterize
        pixel_shapes = transform_geometries_to_pixel_coordinates(
            shape_labels[key], window_bounds
        )

        label_mask = rasterize_shapes_to_mask(
            width, height, pixel_shapes, nodata_value, dtype
        )
        raster_labels.append(RasterLabel(key=key, value=label_mask))

    return raster_labels


def project_geometry_to_crs(
    geometry: BaseGeometry, window_resolution: float, crs: str
) -> STGeometry:
    """Project a geometry to a specified coordinate system.

    Args:
        geometry: The geometry to project
        window_resolution: Resolution in meters per pixel
        crs: The target coordinate reference system

    Returns:
        STGeometry projected to the specified CRS
    """
    # Create source geometry in WGS84
    src_geometry = STGeometry(WGS84_PROJECTION, geometry, None)

    # Create projection with specified resolution
    destination_projection = Projection(
        crs,
        window_resolution,
        -window_resolution,
    )

    # Project the geometry
    return src_geometry.to_projection(destination_projection)



from typing import assert_never, cast

from rslearn.config import dataset

from olmoearth_run.runner.models.operational_context import DatasetOperationalContext
from olmoearth_run.runner.tools.olmoearth_config.constants import LABEL_LAYER_NAME, OUTPUT_LAYER_NAME
from olmoearth_run.runner.tools.olmoearth_config.utils import OutputFieldUtils
from olmoearth_shared.models.olmoearth_config.data.output import RasterField, RasterOutput, SegmentationField, RegressionField, VectorOutput
from olmoearth_shared.models.olmoearth_config.olmoearth_config import OlmoEarthConfig


class LabelOutputLayers:
    @staticmethod
    def generate_label_and_output_layers(olmoearth_config: OlmoEarthConfig, ops_context: DatasetOperationalContext) -> dict[str, dataset.LayerConfig]:
        """Generates all required rslearn label and output layers from the unified config, based on the output schemas and model task definitions."""
        label_output_layers: dict[str, dataset.LayerConfig] = {}

        output = olmoearth_config.data.output
        model = olmoearth_config.model

        if output is None:
            raise ValueError("No output schema specified")

        if len(output.fields) != 1:
            raise ValueError("Only one output field per model is supported at this time")

        field_name = list(output.fields.keys())[0]
        field = list(output.fields.values())[0]

        match output:
            case RasterOutput():
                label_output_layers[LABEL_LAYER_NAME] = LabelOutputLayers.generate_layer_for_raster_field(field_name, cast(RasterField, field), olmoearth_config, ops_context)
                if model:
                    label_output_layers[OUTPUT_LAYER_NAME] = LabelOutputLayers.generate_layer_for_raster_field(field_name, cast(RasterField, field), olmoearth_config, ops_context)
            case VectorOutput():
                raise ValueError("Vector outputs are not supported at this time")
            case _ as unreachable:
                assert_never(unreachable)

        return label_output_layers

    @staticmethod
    def generate_layer_for_raster_field(field_name: str, field: RasterField, olmoearth_config: OlmoEarthConfig, ops_context: DatasetOperationalContext) -> dataset.LayerConfig:
        """These may diverge for label vs output layers in the future, but are currently identical."""
        match field:
            case SegmentationField():
                return RasterFieldTranspilers.generate_segmentation_layer(field_name, field)
            case RegressionField():
                raise ValueError("Regression fields are not supported at this time")
            case _:
                assert_never(field)


class RasterFieldTranspilers:
    """Transpilers for individual raster field types into rslearn layers."""

    @staticmethod
    def generate_segmentation_layer(field_name: str, field: SegmentationField) -> dataset.LayerConfig:
        dtype = OutputFieldUtils.dtype_from_allowed_values(field.allowed_values, field.nodata_value)

        return dataset.LayerConfig(
            type=dataset.LayerType.RASTER,
            band_sets=[
                dataset.BandSetConfig(
                    bands=[field_name],
                    dtype=dtype,
                )
            ],
        )



#TODO(chrisw): this is going away at some point
from enum import StrEnum
from typing import Any

from pydantic import BaseModel, Field, field_validator


class InferenceResultsDataType(StrEnum):
    RASTER = "RASTER"
    VECTOR = "VECTOR"


class PredictionValue(BaseModel):
    """this is a single class value that the model can choose from on a given ClassificationField"""
    value: int | str = Field(description="The computer-friendly value stored in the raster or geojson")
    label: str = Field(description="The human-readable label for this value")
    color: tuple[int, int, int] | tuple[int, int, int, int] = Field(description="The RGB(A) color for this value")

    @field_validator("color", mode="after")
    @classmethod
    def validate_color_values(cls, v: tuple[int, int, int]) -> tuple[int, int, int]:
        """Validate that each RGB value is between 0 and 255."""
        for _, val in enumerate(v):
            if not 0 <= val <= 255:
                raise ValueError(f"Color value must be between 0 and 255, got {val}")
        return v


class ClassificationField(BaseModel):
    """A field the model will try to classify"""
    property_name: str = Field(
        description="The property field name in the geojson feature, or just the name of this band in a raster file"
    )
    confidence_property_name: str | None = Field(
        default=None,
        description="For vector results, if this field has a corresponding confidence field, the name of that field"
    )
    allowed_values: list[PredictionValue]
    band_index: int | None = Field(
        default=None,
        description="For raster results, the index of this field's band in the raster (1-based)"
    )
    confidence_band_index: int | None = Field(
        default=None,
        description="For raster results, if this field has a corresponding confidence value, the band of that field"
    )

    def build_colormap(self) -> dict[int, tuple[int, int, int, int]]:
        """Generates a rio_tiler compliant colormap and adds in alpha if missing"""
        return {
            int(av.value): av.color if len(av.color) == 4 else av.color + (255,)
            for av in self.allowed_values
        }


class RegressionField(BaseModel):
    """A field that the model tries to predict a continuous value of"""
    property_name: str = Field(
        description="The property field name in the geojson feature, or just the name of this band in a raster file")
    band_index: int | None = Field(
        default=None,
        description="For raster results, the index of this field's band in the raster (1-based)")
    min_value: float = Field(description="Minimum possible value for this field")
    max_value: float = Field(description="Maximum possible value for this field")
    colormap_name: str = Field(
        default="viridis",
        description="Name of the rio_tiler colormap to use, "
                    "see https://cogeotiff.github.io/rio-tiler/colormap/#default-rio-tilers-colormaps"
    )


class DetectionObject(BaseModel):
    """For detection models; this is an object the model is trying to detect"""
    detected_object_name: str = Field(description="The name of the object the model is trying to detect (e.g., 'car')")
    confidence_property_name: str | None = Field(
        default=None,
        description="For vector results, if there is a corresponding confidence field, this is the name of that field")
    confidence_band_index: int | None = Field(
        default=None,
        description="For raster results, if there is a corresponding confidence value, this the band of that field"
    )


class InferenceResultsConfig(BaseModel):
    """Configuration for inference results."""
    data_type: InferenceResultsDataType = Field(
        description="The type of predictions that this model returns. Options: RASTER, VECTOR"
    )
    classification_fields: list[ClassificationField] | None = Field(
        default=None,
        description="Classification legend mapping pixel values to labels and colors for raster outputs"
    )
    regression_fields: list[RegressionField] | None = Field(
        default=None,
        description='Fields that the model will predict a continous value on'
    )
    detection_objects: list[DetectionObject] | None = Field(
        default=None,
        description='For a detection model, this is the object(s) that the model is trying to find'
    )

    @field_validator("data_type", mode="before")
    @classmethod
    def uppercase_data_type(cls, v: Any) -> Any:
        if isinstance(v, str):
            return v.upper()
        return v



from geojson_pydantic import FeatureCollection
from shapely.geometry import shape

from olmoearth_shared.api.run.prediction_geometry import PredictionResultCollection
from olmoearth_run.runner.tools.postprocessors.postprocess_interface import PostprocessInterfaceVector


class CombineAndDeduplicateGeojson(PostprocessInterfaceVector):
    def __init__(self, min_iou: float = 0.9, point_nearness: float = 0.001) -> None:
        """
        :param min_iou: Minimum Intersection over Union (IoU) threshold for deduplication.
        :param point_nearness: Distance threshold for considering points as duplicates.
        """
        self.min_iou = min_iou
        self.point_nearness = point_nearness

    def process_partition(self, all_window_results: list[PredictionResultCollection]) -> PredictionResultCollection:
        # Combine all features
        all_features = []
        for collection in all_window_results:
            all_features.extend(collection.features)

        unique_features = []
        processed_indices = set()

        for i, feature in enumerate(all_features):
            if i in processed_indices:
                continue
            if not feature.geometry:
                continue

            geom = shape(feature.geometry.model_dump())
            geom_type = feature.geometry.type

            # Find duplicates for this feature
            duplicates = [i]

            for j, other_feature in enumerate(all_features[i + 1 :], i + 1):
                if j in processed_indices:
                    continue
                if not other_feature.geometry:
                    continue

                other_geom = shape(other_feature.geometry.model_dump())
                other_type = other_feature.geometry.type

                is_duplicate = False

                # Check polygon duplicates using IoU
                if geom_type in ["Polygon", "MultiPolygon"] and other_type in ["Polygon", "MultiPolygon"]:
                    if geom.is_valid and other_geom.is_valid:
                        intersection_area = geom.intersection(other_geom).area
                        union_area = geom.union(other_geom).area

                        if union_area > 0:
                            iou = intersection_area / union_area
                            if iou >= self.min_iou:
                                is_duplicate = True

                # Check point duplicates using distance
                elif geom_type == "Point" and other_type == "Point":
                    distance = geom.distance(other_geom)
                    if distance <= self.point_nearness:
                        is_duplicate = True

                # Check other geometry types for exact matches
                elif geom_type == other_type:
                    if geom.equals(other_geom):
                        is_duplicate = True

                if is_duplicate:
                    duplicates.append(j)

            # Keep the first occurrence, mark others as processed
            unique_features.append(feature)
            processed_indices.update(duplicates)

        return FeatureCollection(type="FeatureCollection", features=unique_features)



from upath import UPath


from olmoearth_shared.models.olmoearth_config.olmoearth_config import OlmoEarthConfig
from olmoearth_run.shared.models.olmoearth_run_config import OlmoEarthRunConfig
from olmoearth_run.runner.tools.olmoearth_run_config_loader import OlmoEarthRunConfigLoader


class ConfigClassLoader:
    @staticmethod
    def load_olmoearth_run_config(config_path: str) -> OlmoEarthRunConfig:
        """Loads legacy OlmoEarthRunConfig from a YAML file into pydantic runtime object"""
        return OlmoEarthRunConfigLoader.load_olmoearth_run_config(config_path)

    @staticmethod
    def load_olmoearth_config(config_path: str) -> OlmoEarthConfig:
        """Loads unified OlmoEarthConfig from a YAML file into pydantic runtime object"""
        with UPath(config_path).open('r') as file:
            yaml_content = file.read()
        return OlmoEarthConfig.from_yaml(yaml_content)



from typing import Literal, TypeVar
from uuid import UUID

from pydantic import BaseModel, Field

from olmoearth_run.shared.models.api.step_type import StepType

###############
# Task Args
###############


class _CommonInferenceTaskArgs(BaseModel):
    # all tasks in the inference workflow get the following fields:
    step_type: StepType
    scratch_path: str = Field(description="A path in a cloud blob storage where the worker can write data to")
    model_stage_id: UUID = Field(description="The ID of the model stage that is being executed")
    model_stage_root_path: str = Field(description="A path in cloud blob storage that contains the configuration files for the model stage that is being executed")
    dataset_path: str = Field(description="Path in cloud blob storage to the dataset for this model stage")

    def get_name_for_id(self) -> str:
        """
        Task Ids are deterministically generated by their step id + a name. This method provides the name part
        It should uniquely identify this piece of work within a single step, and is used to prevent duplicate tasks.
        """
        raise NotImplementedError


class PrepareLabeledWindowsTaskArgs(BaseModel):
    step_type: Literal[StepType.PREPARE_LABELED_WINDOWS] = Field(default=StepType.PREPARE_LABELED_WINDOWS)
    dataset_path: str = Field(description="Path in cloud blob storage to the training dataset for this fine-tuned model")
    scratch_path: str = Field(description="A path in a cloud blob storage where the worker can write data to")

    def get_name_for_id(self) -> str:
        return str(StepType.PREPARE_LABELED_WINDOWS)


class FineTuneTaskArgs(BaseModel):
    step_type: Literal[StepType.FINE_TUNE] = Field(default=StepType.FINE_TUNE)
    dataset_path: str = Field(description="Path in cloud blob storage to the training dataset for this fine-tuned model")
    scratch_path: str = Field(description="A path in cloud blob storage to the fine-tuning scratch space")
    step_id: UUID = Field(description="The ID of the fine-tuning step running this task")

    def get_name_for_id(self) -> str:
        return str(StepType.FINE_TUNE)


class ModelEvaluationTaskArgs(BaseModel):
    step_type: Literal[StepType.MODEL_EVALUATION] = Field(default=StepType.MODEL_EVALUATION)
    dataset_path: str = Field(description="Path in cloud blob storage to the dataset for this fine-tuned model")
    scratch_path: str = Field(description="A path in cloud blob storage to the fine-tuning scratch space")
    checkpoint_path: str = Field(description="Path to the checkpoint file to use for evaluation")

    def get_name_for_id(self) -> str:
        return str(StepType.MODEL_EVALUATION)


class CreatePartitionsTaskArgs(_CommonInferenceTaskArgs):
    step_type: Literal[StepType.CREATE_PARTITIONS] = Field(default=StepType.CREATE_PARTITIONS)
    mask_gcs_path: str | None = Field(default=None, description="GCS path to mask file or directory")
    mask_valid_values: list[int] = Field(default=[1], description="Pixel values considered valid for inference")

    def get_name_for_id(self) -> str:
        return str(StepType.CREATE_PARTITIONS)


class DatasetBuildTaskArgs(_CommonInferenceTaskArgs):
    step_type: Literal[StepType.DATASET_BUILD] = Field(default=StepType.DATASET_BUILD)
    partition_ids: list[str]
    min_window_success_rate: float | None = Field(default=None, description="Minimum required ratio of non-rejected windows (prepared + skipped) to total windows. If None, the check is skipped.")

    def get_name_for_id(self) -> str:
        return "::".join(sorted(self.partition_ids))


class DatasetBuildFromWindowsTaskArgs(BaseModel):
    step_type: Literal[StepType.DATASET_BUILD_FROM_WINDOWS] = Field(default=StepType.DATASET_BUILD_FROM_WINDOWS)
    dataset_path: str = Field(description="Path in cloud blob storage to the dataset for this model stage")
    worker_index: int = Field(default=0, description="Index of this worker (0-based)")
    total_workers: int = Field(default=1, description="Total number of parallel workers")
    min_window_success_rate: float | None = Field(default=None, description="Minimum required ratio of non-rejected windows (prepared + skipped) to total windows. If None, the check is skipped.")
    olmoearth_config_path: str | None = Field(default=None, description="Path to OlmoEarthConfig JSON file for JIT transpilation. If None, expects legacy config.json at dataset_path.")

    def get_name_for_id(self) -> str:
        return f"{StepType.DATASET_BUILD_FROM_WINDOWS}_{self.worker_index}"


class RunInferenceTaskArgs(_CommonInferenceTaskArgs):
    step_type: Literal[StepType.RUN_INFERENCE] = Field(default=StepType.RUN_INFERENCE)
    partition_ids: list[str]

    def get_name_for_id(self) -> str:
        return "::".join(sorted(self.partition_ids))


class PostprocessPartitionTaskArgs(_CommonInferenceTaskArgs):
    step_type: Literal[StepType.POSTPROCESS_PARTITION] = Field(default=StepType.POSTPROCESS_PARTITION)
    partition_ids: list[str]
    mask_gcs_path: str | None = Field(default=None, description="GCS path to mask file or directory")
    mask_valid_values: list[int] = Field(default=[1], description="Pixel values considered valid for inference")

    def get_name_for_id(self) -> str:
        return "::".join(sorted(self.partition_ids))


class CombinePartitionsTaskArgs(_CommonInferenceTaskArgs):
    step_type: Literal[StepType.COMBINE_PARTITIONS] = Field(default=StepType.COMBINE_PARTITIONS)
    partition_ids: list[str]

    def get_name_for_id(self) -> str:
        return str(StepType.COMBINE_PARTITIONS)


TaskArgs = (CombinePartitionsTaskArgs | CreatePartitionsTaskArgs | DatasetBuildTaskArgs |
            DatasetBuildFromWindowsTaskArgs | FineTuneTaskArgs | PostprocessPartitionTaskArgs |
            PrepareLabeledWindowsTaskArgs | RunInferenceTaskArgs | ModelEvaluationTaskArgs)
TaskArgsType = TypeVar("TaskArgsType", bound=TaskArgs)



"""
Retry configuration for different step types.
Defines automatic retry behavior with exponential backoff.
"""
import random

from pydantic import BaseModel, Field

from olmoearth_run.shared.models.api.step_type import StepType


class StepRetryConfig(BaseModel):

    max_retries: int = Field(
        default=0,
        ge=0,
        description="Maximum number of retries after initial failure. 0 means no retries."
    )
    base_delay_seconds: float = Field(
        default=60.0,
        ge=0,
        description="Delay in seconds between retries. Actual delay = base_delay * (2 ^ retry_attempt)"
    )

    def get_delay_seconds(self, retry_attempt: int) -> float:
        """
        Number of seconds to wait before the next retry attempt using exponential backoff with jitter
        This picks a random delay between half the delay and the total delay. Helps to avoid herd risk.
        """
        constant_delay = self.base_delay_seconds * (2 ** retry_attempt)
        return (constant_delay) / 2 + random.uniform(0, (constant_delay) / 2)


# Same defaults on all step types to start with to keep it simple
# These should be tuned per step type based on observations in real data
# NOTE: If the step type is not in the dictionary, the automatic retries are disabled
STEP_TYPE_RETRY_CONFIGS: dict[StepType, StepRetryConfig] = {
    StepType.PREPARE_LABELED_WINDOWS: StepRetryConfig(max_retries=3, base_delay_seconds=60.0),
    StepType.CREATE_PARTITIONS: StepRetryConfig(max_retries=2, base_delay_seconds=60.0),
    StepType.DATASET_BUILD: StepRetryConfig(max_retries=5, base_delay_seconds=60.0),
    StepType.DATASET_BUILD_FROM_WINDOWS: StepRetryConfig(max_retries=3, base_delay_seconds=60.0),
    StepType.FINE_TUNE: StepRetryConfig(max_retries=3, base_delay_seconds=60.0),
    StepType.RUN_INFERENCE: StepRetryConfig(max_retries=2, base_delay_seconds=60.0),
    StepType.POSTPROCESS_PARTITION: StepRetryConfig(max_retries=2, base_delay_seconds=60.0),
    StepType.COMBINE_PARTITIONS: StepRetryConfig(max_retries=2, base_delay_seconds=60.0),
}


def get_retry_config(step_type: StepType) -> StepRetryConfig:
    return STEP_TYPE_RETRY_CONFIGS.get(step_type, StepRetryConfig())



from enum import StrEnum
from typing import Generic, TypeAlias, TypeVar
from uuid import UUID

import httpx
import requests
from pydantic import BaseModel, ConfigDict, Field, ValidationError
from typing_extensions import deprecated

T = TypeVar("T")
HttpResponse: TypeAlias = httpx.Response | requests.Response


class Metadata(BaseModel):
    model_config = ConfigDict(extra="forbid")
    total: int | None = Field(default=None)


class ApiErrorCode(StrEnum):
    NOT_FOUND_ERROR = "not_found_error"
    PERMISSION_ERROR = "permission_error"
    SERVER_ERROR = "server_error"
    UNAUTHORIZED_ERROR = "unauthorized_error"
    VALIDATION_ERROR = "validation_error"
    NOT_IMPLEMENTED_ERROR = "not_implemented_error"


class ApiOutputError(BaseModel):
    code: ApiErrorCode
    message: str


class ApiOutputModel(BaseModel, Generic[T]):
    records: list[T] | None = Field(default=None)
    meta: Metadata | None = Field(default=None)
    errors: list[ApiOutputError] | None = Field(default=None)

    @classmethod
    def from_http_response(cls: type["ApiOutputModel[T]"], response: HttpResponse) -> "ApiOutputModel[T]":
        """
        Helper method to parse an HTTP response into this class
        Handles errors by raising an exception that includes those errors.
        """
        if isinstance(response, httpx.Response):
            return cls._from_httpx_response(response)
        if isinstance(response, requests.Response):
            return cls._from_requests_response(response)
        raise TypeError(f"Unsupported response type: {type(response)}")

    @classmethod
    def _from_httpx_response(cls: type["ApiOutputModel[T]"], response: httpx.Response) -> "ApiOutputModel[T]":
        url = response.request.url
        try:
            output = cls(**response.json())
        except ValidationError as e:
            # raise Exception(f"Failed to parse response from {url}: {response.text[0:1000]}") from e  # Revert to me
            raise Exception(f"Failed to parse response from {url}: {response.text}") from e
        except Exception as e:
            raise Exception(f"Unexpected error handling response from {url}: {response.text[0:1000]}") from e

        if not response.is_success:
            message = f"Received HTTP/{response.status_code} from {url} - {response.text}"
            raise ApiError(message, output, response.status_code)
        return output

    @classmethod
    def _from_requests_response(cls: type["ApiOutputModel[T]"], response: requests.Response) -> "ApiOutputModel[T]":
        url = response.url
        try:
            output = cls(**response.json())
        except ValidationError as e:
            raise Exception(f"Failed to parse response from {url}: {response.text}") from e
        except Exception as e:
            raise Exception(f"Unexpected error handling response from {url}: {(response.text or '')[0:1000]}") from e

        if not response.ok:
            message = f"Received HTTP/{response.status_code} from {url} - {response.text}"
            raise ApiError(message, output, response.status_code)
        return output

    @classmethod
    def from_pydantic_validation_error(cls: type['ApiOutputModel[T]'], e: ValidationError) -> 'ApiOutputModel[T]':
        """
        Helper method to create an ApiOutputModel from a Pydantic ValidationError
        """
        errors = [
            ApiOutputError(code=ApiErrorCode.VALIDATION_ERROR,
                           message=f"Validation Error for {e.title}: {err['msg']} "
                                   f"[type={err['type']}, input_value={err['input']}, location={err['loc']}]")
            for err in e.errors()
        ]
        return cls(errors=errors)


@deprecated("Use ApiOutputModel[DeleteResource] instead")
class DeleteResource(BaseModel):
    """
    A response from a delete request
    """
    id: UUID
    deleted: bool


class ApiDeletionStatus(StrEnum):
    DELETED = "deleted"
    """Immediate deletion."""

    DELETION_QUEUED = "deletion_queued"
    """Queued for deletion."""

    REMOVED = "removed"
    """Removed as child of a parent record without deletion."""


class DeletionRead(BaseModel, Generic[T]):
    record: T = Field(..., description="The record targeted for deletion")
    status: ApiDeletionStatus = Field(
        ...,
        description=(
            "Indicates whether record was deleted immediately, queued for deletion, "
            "or removed as child of a parent without deletion."
        ),
    )


ApiDeletionModel: TypeAlias = ApiOutputModel[DeletionRead[T]]


class ApiError(Exception, Generic[T]):
    """
    An exception that can include the errors that were returned by an ApiOutputModel
    """

    errors: list[ApiOutputError]

    def __init__(self, message: str, api_output: ApiOutputModel[T], http_status_code: int) -> None:
        super().__init__(message)
        self.errors = api_output.errors or []
        self.http_status_code = http_status_code
